(* +------------------------------------------------------------------------+
   |                                                                        |
   |                     Paralllisation d'un calcul                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* M. Quercia, mars 1999 *)

open Printf
open Unix

exception Job_wait
exception Job_giveup

(* messages envoys  un esclave *)
type ('init,'job) m_message =
  | INIT of 'init
  | JOB  of 'job
  | GIVEUP
  | FINISHED

(* messages envoys  un matre *)
type 'result s_message =
  | RESULT of 'result
  | ABORTED
  | DEAD


                            (* +-------------+
                               |  Dboguage  |
                               +-------------+ *)

let debug = ref(false)

let log(message) =
  let t = gettimeofday() in
  let {tm_hour=h; tm_min=m; tm_sec=s} = localtime(t)
  and frac = Pervasives.truncate(100.0*.(t -. floor(t)))
  in
  fprintf Pervasives.stderr "%02d:%02d:%02d.%02d %s\n" h m s frac message;
  flush Pervasives.stderr


               (* +---------------------------------+
                  |  Dcode une chaine "host:port"  |
                  +---------------------------------+ *)

let dcode(chaine) =
  let l = String.length(chaine) in
  let i = ref(l-1) in
  while !i >= 0 & chaine.[!i] <> ':' do decr i done;
  let port = int_of_string(String.sub chaine (!i+1) (l - !i - 1))
  and host = (if !i > 0 then String.sub chaine 0 !i else "localhost") in
  let addr =
    try inet_addr_of_string host
    with Failure _ ->
      try (gethostbyname host).h_addr_list.(0)
      with Not_found -> failwith (sprintf "unknown host %s" host) in
  (addr,port)


            (* +---------------------------------------+
               |  Tentatives d'ouverture d'une socket  |
               +---------------------------------------+ *)

let tente f addr =
  let rec essai(n) =
    let s = socket PF_INET SOCK_STREAM 0 in
    try
      f s addr;
      s
    with ex ->
      close s;
      if n = 0 then raise ex else (sleep 5; essai(n-1))
  in essai(5)


module SafeIo = struct

(*
 * Il y a un problme avec le module io : lorsque plusieurs donnes arrivent
 * en squence, Unix.select dtecte la premire donne mais pas les suivantes.
 * On rimplmente les canaux d'entre pour contourner le problme.
 *)

type in_channel = {
  d : Unix.file_descr;      (* descripteur unix      *)
  mutable buff : string;    (* tampon de rception   *)
  mutable i    : int;       (* indice de remplissage *)
  mutable l    : int;       (* nb d'octets attendus  *)
  mutable err  : bool       (* read error            *)
}

(* liste des canaux ouverts *)
let chlist = ref []

(* remplacement des fonctions standard *)
let descr_of_in_channel ch = ch.d

let in_channel_of_descr d  =
  let ch = {d=d; buff = String.create 100; i=0; l = -1; err = false} in
  chlist := ch :: !chlist;
  ch

let close_in ch =
  let rec remove = function
    | [] -> []
    | c::suite when c == ch -> suite
    | c::suite -> c :: remove suite in
  chlist := remove !chlist;
  try Unix.close ch.d with Unix_error(_,"close",_) -> ()

(* est-ce qu'on peut lancer input_value ? *)
let is_ready ch = (ch.l >= 0 & ch.i >= ch.l) or ch.err

(* rception de donnes *)
let read_in ch =
  try
    let i = ch.i
    and j = ch.i + Unix.read ch.d ch.buff ch.i (String.length ch.buff - ch.i) in
    ch.i <- j;
    if ch.l < 0 & j >= Marshal.header_size then begin
      ch.l <- Marshal.total_size ch.buff 0;
      if ch.l > String.length ch.buff then begin
	let s = String.create ch.l in
	String.blit ch.buff 0 s 0 ch.i;
	ch.buff <- s
      end
    end
  with Unix_error(_,"read",_) -> ch.err <- true

(* remplacement de input_value *)
let input_value ch =
  if ch.l >= 0 & ch.i >= ch.l then begin
    let v = Marshal.from_string ch.buff 0 in
    String.blit ch.buff ch.l ch.buff 0 (ch.i - ch.l);
    ch.i <- ch.i - ch.l;
    if ch.i >= Marshal.header_size then begin
      ch.l <- Marshal.total_size ch.buff 0;
      if ch.l > String.length ch.buff then begin
	let s = String.create ch.l in
	String.blit ch.buff 0 s 0 ch.i;
	ch.buff <- s
      end
    end
    else ch.l <- -1;
    v
  end
  else failwith "input_value"

(* liste des descripteurs sur lequels on peut faire input_value *)
let rec search_ready fds = function
  | [] -> fds
  | ch::suite -> if is_ready ch then search_ready (ch.d::fds) suite
    else search_ready fds suite

(* cherche si un descripteur correspond  in in_channel *)
let rec find_channel d = function
  | [] -> raise Not_found
  | ch::_ when ch.d = d -> ch
  | _::suite -> find_channel d suite

(* dcoupe une liste de descripteurs en in_channels et autres *)
let rec split chans fds = function
  | [] -> chans,fds
  | d::suite -> try let ch = find_channel d !chlist in
                    split (ch::chans) fds suite
                with Not_found -> split chans (d::fds) suite

(* rimplementation de select *)
let select inlist outlist errlist timeout =

  let chinlist,_ = split [] [] inlist in
  match search_ready [] chinlist with

    | [] -> begin
	let rec loop() =
	  try
	    let inlist,outlist,errlist = Unix.select inlist outlist errlist timeout in
	    let chans,fds = split [] [] inlist in
	    List.iter read_in chans;
	    let inlist = (search_ready [] chans) @ fds in
	    if inlist <> [] or outlist <> [] or errlist <> []
	    then (inlist,outlist,errlist)
	    else loop()
	  with Unix.Unix_error(_,"select",_) -> loop() in
	loop()
      end

    | l -> l,[],[]

end (* module SafeIO *)	    

(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                Esclave                                 |
   |                                                                        |
   +------------------------------------------------------------------------+ *)


module Slave = struct

                   (* +-------------------------+
                      |  Structures de donnes  |
                      +-------------------------+ *)

  type ('job,'result) tat =
    | Job    of 'job
    | Computing
    | Result of 'result
    | Aborting
    | Waiting
    | Finished

  type ('init,'job,'result) slave = {
    mutable tat : ('job,'result) tat;
    chin  : in_channel;     	   (* matre -> esclave 	*)
    chout : out_channel;    	   (* esclave -> matre 	*)
    soi   : 'init   -> string;     (* fonctions d'impression    *)
    soj   : 'job    -> string;
    sor   : 'result -> string
  }


                 (* +-----------------------------+
                    |  Trace les entres-sorties  |
                    +-----------------------------+ *)

  (* envoie un rsultat *)
  let soutput s r = try
    if !debug then log(sprintf "< %s" (match r with
       | RESULT(r) -> "RESULT " ^ (s.sor r)
       | ABORTED   -> "ABORTED"
       | DEAD      -> "DEAD"));
    output_value s.chout r;
    flush s.chout
  with ex ->
    if !debug then log "output_value failed";
    s.tat <- Finished;
    raise ex

  (* reoit un message *)
  let sinput s = try
    match s.tat with
      | Finished -> FINISHED
      |_ -> let v = input_value(s.chin) in
	    if !debug then log(sprintf "> %s" (match v with
	       | INIT(i) ->  "INIT " ^ (s.soi i)
	       | JOB(j)  ->  "JOB "  ^ (s.soj j)
	       | GIVEUP  ->  "GIVEUP"
	       | FINISHED -> "FINISHED"));
	    v
  with ex ->
    if !debug then log "input_value failed";
    s.tat <- Finished;
    raise ex

 (* teste si une demande d'abandon est arrive *)
  let try_giveup s =
    match s.tat with
      | (Aborting | Finished) -> raise Job_giveup
      | Computing ->
	  let a,_,_ = select [descr_of_in_channel s.chin] [] [] 0.0 in
	  if a <> [] then begin
	    match sinput s with
	      | GIVEUP   -> s.tat <- Aborting; raise Job_giveup
	      | FINISHED -> s.tat <- Finished; raise Job_giveup
	      | _ -> if !debug then log "unexpected message"
	  end
      | _ -> ()


                     (* +---------------------+
                        |  Calcul et rapport  |
                        +---------------------+ *)

  (* lance un calcul *)
  let calcul s calc j =
    begin
      s.tat <- Computing;
      try
	let r = calc(j) in
	if s.tat = Computing then s.tat <- Result(r);
      with Job_giveup -> ()
	| ex -> s.tat <- Finished; raise ex
    end;
    match s.tat with
      | Result(r) -> soutput s (RESULT r); s.tat <- Waiting
      | Aborting  -> soutput s ABORTED;    s.tat <- Waiting
      | _ ->	     ()


                     (* +---------------------+
                        |  Boucle de travail  |
                        +---------------------+ *)

  let work s calc =
    try
      while s.tat <> Finished do
	match sinput s with
	  | JOB(j)   -> calcul s calc j;
	  | GIVEUP   -> soutput s ABORTED
	  | FINISHED -> s.tat <- Finished
	  | _        -> if !debug then log "unexpected message"
      done;
      close_in  s.chin;
      close_out s.chout
    with ex ->
      close_in  s.chin;
      close_out s.chout;
      raise ex


                 (* +----------------------------+
                    |  Connexion avec le matre  |
                    +----------------------------+ *)

  (* dmarre les calculs *)
  let init chin chout soi soj sor make_calc =
    let s = {tat  = Waiting;
	     chin  = chin;
	     chout = chout;
	     soi = soi; soj = soj; sor = sor} in
    match sinput s with
      | INIT(i) -> work s (make_calc i s)
      | _       -> failwith "no init data"

  (* contacte le matre *)
  let connect maitre =
    let (h,p) = dcode maitre in
    let sock = tente Unix.connect (ADDR_INET(h,p)) in
    init (in_channel_of_descr sock) (out_channel_of_descr sock)

  (* serveur autonome *)
  let listen port soi soj sor make_calc =
    establish_server
      (fun i o -> init i o soi soj sor make_calc; exit 0)
      (ADDR_INET(inet_addr_of_string("0.0.0.0"), port))

  (* serveur lanc par inetd *)
  let inetd soi soj sor make_calc =
    init (Pervasives.stdin) (Pervasives.stdout) soi soj sor make_calc 

end (* module Slave *)


(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                 Matre                                 |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

module Master = struct
  open SafeIo


                        (* +-------------------------+
                           |  Structures de donnes  |
                           +-------------------------+ *)

  (* tats du calcul parallle *)
  type tat = Continue | Redo | Wait
  type 'job mtat = Job of 'job | Aborting | Waiting

  (* donnes associes  une machine *)
  type 'job machine = {
    addr  : string;     	(* adresse TCP *)
    chin  : in_channel;         (* rception   *)
    chout : out_channel;        (* mission    *)
    mutable j : 'job mtat      (* activit    *)
  }

  type 'job machine_of_descr =
    | Sock    of file_descr
    | Machine of 'job machine
    | Unknown

  (* donnes associes au matre *)
  type ('init,'job,'result) master = {
    init   : 'init;                                     (* init. esclave     *)
    job    : unit -> 'job;                              (* donne un travail  *)
    result : 'job -> 'result -> 'job list -> 'job list; (* rsultat          *)
    mutable calc     : ('job -> 'result) option;        (* calcul autonome   *)
    mutable tat     : tat;                            (* tat du calcul    *)
    mutable machines : 'job machine list;               (* esclaves          *)
    mutable jobs     : 'job list;                       (* travaux en cours  *)
    mutable lost     : 'job list;                       (* travaux  refaire *)
    mutable socks    : file_descr list;                 (* prises TCP        *)
    soi     : 'init   -> string;                   (* fonctions d'impression *)
    soj     : 'job    -> string;
    sor     : 'result -> string
  }


  (* dboguage *)
  let mlog machine msg =
    log(sprintf "%s %s" machine.addr msg)

                 (* +-----------------------------+
                    |  Entres-sorties protges  |
                    +-----------------------------+ *)

  (* supprime une machine *)
  let remove master machine =
    let rec rem = function
      | [] -> if !debug then log "machine to remove not found"; []
      | m::suite when m == machine ->
	  if !debug then log(sprintf "removing %s" m.addr); 
	  close_in m.chin;
	  close_out m.chout;
	  suite
      | m::suite -> m :: rem suite
    in
    master.machines <- rem master.machines;
    match machine.j with
      | Job(j) -> master.lost <- master.lost @ [j]
      | _      -> ()

  (* envoie un message *)
  let moutput master machine data =
    try
      if !debug then mlog machine (match data with
	| INIT(i)  ->  "< INIT " ^ (master.soi i)
	| JOB(j)   ->  "< JOB "  ^ (master.soj  j)
	| GIVEUP   ->  "< GIVEUP"
	| FINISHED ->  "< FINISHED");
      output_value machine.chout data;
      flush machine.chout;
    with _ ->
      if !debug then mlog machine "< output_value failed";
      remove master machine
	  
  (* reoit un message *)
  let minput master machine =
    try
      let v = input_value machine.chin in
      if !debug then mlog machine (match v with
	| RESULT(r) -> "> RESULT " ^ (master.sor r)
	| ABORTED   -> "> ABORTED"
	| DEAD      -> "> DEAD");
      v
    with _ ->
      if !debug then mlog machine "> input_value failed";
      remove master machine;
      DEAD

  (* identifie un descripteur de fichier *)
  let find_descr =
    let rec search_sock d = function
      | [] -> Unknown
      | s::_ when d=s -> Sock(s)
      | _::suite -> search_sock d suite
    and search_machine d = function
      | [] -> Unknown
      | m::_ when d = descr_of_in_channel(m.chin) -> Machine(m)
      | _::suite -> search_machine d suite
    in fun master d -> match search_sock d master.socks with
      | Unknown -> search_machine d master.machines
      | s       -> s



                    (* +-----------------------+
                       |  Gestion des travaux  |
                       +-----------------------+ *)

  (* abandon d'un travail *)
  let mgiveup master j =
    List.iter
      (fun m ->
	 if m.j = Job(j) then begin
	   moutput master m GIVEUP;
	   m.j <- Aborting
	 end)
      master.machines;
    let rec rem = function
      | [] -> []
      | job::suite when job = j -> suite
      | job::suite -> job :: rem suite
    in
    master.jobs <- rem master.jobs;
    master.lost <- rem master.lost

  (* cherche un travail  faire *)
  let rec get_job(master) = match (master.tat,master.lost,master.jobs) with
    | _,j::suite,_     -> master.lost <- suite; j
    | Continue,[],_    -> begin
	try  let j = master.job() in master.jobs <- master.jobs @ [j]; j
	with Job_wait -> master.tat <- Redo; get_job(master)
      end
    | Redo,[],j::suite -> master.jobs <- suite @ [j]; j
    | _ -> raise Job_wait
  
  (* relance les machines inoccupes *)
  let mrestart master =
    let rec go = function
      | [] -> ()
      | m::suite when m.j = Waiting -> begin
	  try
	    let j = get_job(master) in
	    m.j <- Job(j);
	    moutput master m (JOB(j));
	    go suite
	  with Job_wait -> master.tat <- Wait
	end
      | _::suite -> go suite
    in go master.machines


                    (* +-----------------------+
                       |  Gestion des entres  |
                       +-----------------------+ *)


  (* rsultat *)
  let mresult master m = 
    let v = minput master m in match (m.j,v) with
      | _,DEAD             -> ()
      | Aborting,RESULT(_) -> ()
      | Aborting,ABORTED   ->  m.j <- Waiting
      | Job(j),RESULT(r)   ->
	  m.j <- Waiting;
       	  mgiveup master j;
	  let jlist = master.result j r master.jobs in
	  List.iter (mgiveup master) jlist
      | _ ->
	  if !debug then mlog m "> unexpected data";
          remove master m

  (* nouvel esclave *)
  let mslave master d a =
    let chin  = in_channel_of_descr  d 
    and chout = out_channel_of_descr d in
    let m = {
      addr = (match a with
		| ADDR_INET(i,p) -> sprintf "%s:%d" (string_of_inet_addr i) p
		| ADDR_UNIX(s)   -> s);
      chin  = chin;
      chout = chout;
      j     = Waiting} in
    master.machines <- m :: master.machines;
    moutput master m (INIT(master.init))

  (* connexion entrante *)
  let maccept m s =
    try let (d,a) = accept s in mslave m d a
    with _ -> if !debug then log "accept failed"

                         (* +-------------+
                            |  Interface  |
                            +-------------+ *)

  (* initialise la session *)
  let init idata job result soi soj sor = {
    init     = idata;
    job      = job;
    result   = result;
    calc     = None;
    tat     = Continue;
    machines = [];
    jobs     = [];
    lost     = [];
    socks    = [];
    soi = soi; soj = soj; sor = sor}

 (* ouvre un port TCP en coute *)
  let listen master port =
    let s = tente bind (ADDR_INET(inet_addr_of_string "0.0.0.0",port)) in
    Unix.listen s 5;
    master.socks <- s :: master.socks;
    master.calc <- None

  (* contacte un esclave *)
  let connect master machine =
    let (h,p) = dcode machine in
    let sockaddr = ADDR_INET(h,p) in
    let s = tente Unix.connect sockaddr in
    mslave master s sockaddr;
    master.calc <- None

  (* lance les calculs *)
  let start master =
    master.tat <- Continue;
    mrestart master;

    while master.tat <> Wait or master.lost <> [] or master.jobs <> [] do
      match master.calc with

	| Some(calc) -> begin
	    try
	      let j = master.job() in
	      if !debug then log ("JOB " ^ master.soj(j));
	      let r = calc j in
	      if !debug then log ("RESULT " ^ master.sor(r));
	      let _ = master.result j r [] in ()
	    with Job_wait -> master.tat <- Wait
	  end

	| None ->
	    let l = master.socks @
		    (List.map (fun m -> descr_of_in_channel m.chin) master.machines)
	    in
	    let ready,_,error = select l [] l (-1.0) in
	    List.iter	     
	      (fun d -> match find_descr master d with
		 | Machine(m) -> if !debug then mlog m "select error";
		                 remove master m
		 | Sock(_)    -> if !debug then log "socket error"
		 | Unknown    -> ())
	      error;
	    List.iter	     
	      (fun d -> match find_descr master d with
		 | Machine(m) -> mresult master m
		 | Sock(_)    -> maccept master d
		 | Unknown    -> ())
	      ready;
	    mrestart master
    done

  (* arrt de la session *)
  let stop master =
    List.iter
      (fun m ->
	 moutput master m FINISHED;
	 close_out m.chout;
	 close_in m.chin)
      master.machines;
    List.iter close master.socks;
    master.machines <- [];
    master.socks    <- []
    
  (* passe en mode autonome *)
  let autonomous master calc =
    stop master;
    master.calc <- Some(calc)

end (* module Master *)


